@@ -1,11 +1,138 @@
 from typing import Optional
+from typing import Union
 from datetime import datetime
 from datetime import date
 
+from collections.abc import Iterable
 from sqlalchemy.orm.collections import InstrumentedList
 
 from .database import db
 
+CIRCREF_CHECK = set()
+
+
+class Config:
+    collection = False
+
+
+def get_primary_key_value(elem: db.Model):
+    pk = primarykey(elem)
+    return getattr(elem, pk, None)
+
+
+def extract_special_members(data, *args, **kwargs):
+    resp = {}
+    # first extract any special properties or hybrid properties
+    for prop in getattr(data, '__properties__', set()):
+        if not getattr(prop, '__name__', None):
+            resp[prop.__name__] = prop.fget(data)
+        property = serializer(getattr(data, prop.__name__), *args, **kwargs)
+        resp[prop.__name__] = property
+
+    # then get the counts we have defined as special len(relation)
+    if getattr(data, '__counts__', None):
+        resp['__counts__'] = getattr(data, '__counts__')
+
+    # if a Model has some relationships that should be exposed by default, extract them here
+    if getattr(data, '__expose__', None):
+        for entity in getattr(data, '__expose__'):
+            if entity.key in kwargs.get('exclusions', set()):
+                continue
+            if entity.uselist:
+                resp[entity.key] = []
+                for child in getattr(data, entity.key):
+                    resp[entity.key].append(serializer(child, *args, **kwargs))
+            else:
+                resp[entity.key] = serializer(getattr(data, entity.key, None), *args, **kwargs)
+    return resp
+
+
+def extract_relationships(data, **kwargs):
+    if kwargs.get('rels', None) is None:
+        return set()
+    rels = kwargs.get('rels', set())
+    if '*' in rels:
+        rels = relationships(data, fmt=True)
+    resp: dict = {}
+    for rel in set(relationships(data, fmt=True)).intersection(rels):
+        resp[rel] = serializer(getattr(data, rel, None), **kwargs)
+    # CIRCREF_CHECK.clear()
+    return resp
+
+
+def serializer(*args, **kwargs):
+    return _tojson(*args, **kwargs)
+
+
+def tojson(*args, **kwargs):
+    CIRCREF_CHECK.clear()
+    return _tojson(*args, **kwargs)
+
+
+def _tojson(data: Union[list, db.Model, dict], *args, **kwargs):
+    exclude = kwargs.get('exclusions', set())
+    resp: dict = None
+    if isinstance(data, Iterable) and not isinstance(data, (dict, str, tuple)):
+        Config.collection = True
+        transformed = []
+        for (idx, elem) in enumerate(data):
+            if idx == 0:
+                deny_func = getattr(elem, '__denials__', None)
+                if deny_func:
+                    kwargs.get('exclusions').extend(deny_func('__collection__'))
+            transformed.append(_tojson(elem, *args, **kwargs))
+        return transformed
+    elif isinstance(data, db.Model):
+        resp = {}
+        final = set(columns(data, True)).difference(set(exclude))
+        hidden = getattr(data, '__hidden__', set())
+        final = final.difference(hidden)
+        for column in final:
+            resp[column] = getattr(data, column, None)
+
+        if not getattr(data, primarykey(data)) in CIRCREF_CHECK:
+            CIRCREF_CHECK.add(getattr(data, primarykey(data)))
+            resp.update(extract_special_members(data, *args, **kwargs))
+            resp.update(extract_relationships(data, **kwargs))
+
+    elif isinstance(data, tuple):
+        resp = {}
+        # this will cover the cases where you select fields explicitly
+        # and SQLAlchemy returns a results tuple rather than a full db.Model
+        final = set(columns(data, True)).difference(exclude)
+        for column in final:
+            resp[column] = getattr(data, column, None)
+        resp.update(extract_special_members(data, *args, **kwargs))
+    elif isinstance(data, (dict, str, int)):
+        return data
+    # CIRCREF_CHECK.clear()
+    return resp
+
+
+def _serializer(data, exclude=set(), **kwargs):
+    """
+    Takes a SQLAlchemy Model (db.Model) and converts it into a response ready JSON format.
+
+    :param data: list of db.Model types or single instance of db.Model types
+    :param exclude:
+    :return:
+    """
+
+    if not exclude:
+        exclude = kwargs.get('exclusions', set())
+
+    if isinstance(data, db.Model):
+        resp = {}
+        final = set(columns(data, True)).difference(exclude)
+        for column in final:
+            resp[column] = getattr(data, column, None)
+        return resp
+    elif isinstance(data, list):
+        resp = []
+        for elem in data:
+            resp.append(serializer(elem, exclude))
+        return resp
+
 
 def save(session):
     session.save()
@@ -116,7 +243,8 @@ def extract(element, fields=None, exclude: Optional[set] = None, **kwargs) -> Op
             resp[column] = getattr(element, column)
 
     for prop in getattr(element, '__properties__', set()):
-        resp[prop.fget.__name__] = getattr(element, prop.fget.__name__)
+        property = serializer(getattr(element, prop.__name__), exclude)
+        resp[prop.__name__] = property
 
     if getattr(element, '__counts__', None):
         resp['__counts__'] = getattr(element, '__counts__')
@@ -131,7 +259,6 @@ def extract(element, fields=None, exclude: Optional[set] = None, **kwargs) -> Op
                     resp[entity.key].append(extract(child))
             else:
                 resp[entity.key] = extract(getattr(element, entity.key, None))
-
     return resp
 
 
@@ -153,10 +280,12 @@ def transform(data, fields=None, relations=None, **kwargs):
         for item in data:
             resp.append(extract(item))
         return resp
-    return extract(data)
+    elif isinstance(data, dict):
+        return extract(data)
+    return data
 
 
-def iserialize(data, fields=None, rels=None, root=None, exclude=None, functions=None,
+def iserialize(data, fields=None, rels=None, root=None, exclusions=None, functions=None,
                **kwargs):
     """
     This utility function dynamically converts Alchemy model classes into a
@@ -185,10 +314,10 @@ def iserialize(data, fields=None, rels=None, root=None, exclude=None, functions=
 
     if functions is None:
         functions = {}
-    if exclude is None:
+    if exclusions is None:
         exclude = set()
     else:
-        exclude = set(exclude)
+        exclude = set(exclusions)
 
     model = data
     if isinstance(data, list) and len(data):
